@@ -4,6 +4,7 @@
 
 #include <assert.h>
 #include <errno.h>
+#include <limits.h>
 #include <locale.h>
 #include <signal.h>
 #include <stdio.h>
@@ -17,6 +18,11 @@
 
 #include <sodium.h>
 
+#ifdef HAVE_LIBSYSTEMD
+# include <sys/socket.h>
+# include <systemd/sd-daemon.h>
+#endif
+
 #include "app.h"
 #include "dnscrypt_client.h"
 #include "dnscrypt_proxy.h"
@@ -98,6 +104,7 @@ proxy_context_init(ProxyContext * const proxy_context, int argc, char *argv[])
     proxy_context->udp_proxy_resolver_event = NULL;
     proxy_context->udp_proxy_resolver_handle = -1;
     proxy_context->udp_listener_handle = -1;
+    proxy_context->tcp_listener_handle = -1;
     if (options_parse(&app_context, proxy_context, argc, argv) != 0) {
         return -1;
     }
@@ -227,9 +234,10 @@ dnscrypt_proxy_start_listeners(ProxyContext * const proxy_context)
                                 resolver_addr_s, sizeof resolver_addr_s);
     logger(proxy_context, LOG_NOTICE, "Proxying from %s to %s",
            local_addr_s, resolver_addr_s);
-
     proxy_context->listeners_started = 1;
-
+#ifdef HAVE_LIBSYSTEMD
+    sd_notify(0, "READY=1");
+#endif
     return 0;
 }
 
@@ -245,6 +253,55 @@ dnscrypt_proxy_loop_break(void)
     return 0;
 }
 
+#ifdef HAVE_LIBSYSTEMD
+static int
+init_descriptors_from_systemd(ProxyContext * const proxy_context)
+{
+    int num_sd_fds;
+    int sock;
+
+    num_sd_fds = sd_listen_fds(0);
+    if (num_sd_fds == 0) {
+        return 0;
+    }
+    if (num_sd_fds != 2) {
+        logger(proxy_context, LOG_ERR, "Wrong number of systemd sockets: %d - "
+               "should be 2", num_sd_fds);
+        return -1;
+    }
+    assert(num_sd_fds <= INT_MAX - SD_LISTEN_FDS_START);
+    for (sock = SD_LISTEN_FDS_START; sock < SD_LISTEN_FDS_START + num_sd_fds;
+         ++sock) {
+       if (sd_is_socket(sock, AF_INET, SOCK_DGRAM, 0) > 0 ||
+           sd_is_socket(sock, AF_INET6, SOCK_DGRAM, 0) > 0) {
+           proxy_context->udp_listener_handle = sock;
+       }
+       if (sd_is_socket(sock, AF_INET, SOCK_STREAM, 1) > 0 ||
+           sd_is_socket(sock, AF_INET6, SOCK_STREAM, 1) > 0) {
+           proxy_context->tcp_listener_handle = sock;
+       }
+    }
+    if (proxy_context->udp_listener_handle < 0) {
+        logger_noformat(proxy_context, LOG_ERR,
+                        "No systemd UDP socket passed in");
+        return -1;
+    }
+    if (proxy_context->tcp_listener_handle < 0) {
+        logger_noformat(proxy_context, LOG_ERR,
+                        "No systemd TCP socket passed in");
+        return -1;
+    }
+    if (getsockname(proxy_context->udp_listener_handle,
+                    (struct sockaddr *) &proxy_context->local_sockaddr,
+                    &proxy_context->local_sockaddr_len) != 0) {
+        logger_noformat(proxy_context, LOG_ERR,
+                        "Unable to get the local systemd UDP socket address");
+        return -1;
+    }
+    return 0;
+}
+#endif
+
 int
 dnscrypt_proxy_main(int argc, char *argv[])
 {
@@ -290,6 +347,11 @@ dnscrypt_proxy_main(int argc, char *argv[])
     if (cert_updater_init(&proxy_context) != 0) {
         exit(1);
     }
+#ifdef HAVE_LIBSYSTEMD
+    if (init_descriptors_from_systemd(&proxy_context) != 0) {
+        exit(1);
+    }
+#endif
     if (proxy_context.test_only == 0 &&
         (udp_listener_bind(&proxy_context) != 0 ||
          tcp_listener_bind(&proxy_context) != 0)) {
@@ -303,10 +365,18 @@ dnscrypt_proxy_main(int argc, char *argv[])
     if (cert_updater_start(&proxy_context) != 0) {
         exit(1);
     }
+
+#ifdef HAVE_LIBSYSTEMD
+    sd_notifyf(0, "MAINPID=%lu", (unsigned long) getpid());
+#endif
     if (skip_dispatch == 0) {
         event_base_dispatch(proxy_context.event_loop);
     }
     logger_noformat(&proxy_context, LOG_NOTICE, "Stopping proxy");
+#ifdef HAVE_LIBSYSTEMD
+    sd_notify(0, "STOPPING=1");
+#endif
+
     cert_updater_free(&proxy_context);
     udp_listener_stop(&proxy_context);
     tcp_listener_stop(&proxy_context);
